Option Strict Off
Option Explicit On

#Const FULLBIDASKDATA = 1  ' 0=none  1=Bid/ask  2=OHLC formed from bid and ask, bid OHLC & ask OHLC

Module NSTDataPumpAPI
    ' This moudle contains procedures that can be called to simplify coding the interface to the NeuroShell Trader Data Pump dll

    ' *****************************************************************************************************
    ' SETTINGS
    ' *****************************************************************************************************

    ' Setup the DataServerTypes constant based upon how your data server sends historical and realtime data bars.  (comment out the unused ones)
    '   = 1  Your data server provides OHLC historical data, but only tick by tick realtime updates for minute/hour bars
    '   = 2  Your data server provides OHLC historical data, and not all realime ticks are captured, but instead updated OHLC updates are sent or polled for.
    '   = 0  Your data server provides only tick by tick historical and tick by tick realtime updates for minute/hour bars
    Public Const DataServerType As Short = 2

    'This will inform NeuroShell Trader that a UserID and Password are required
    ' - NeuroShell Trader will prompt the user for the UserID and Password, then inform this server
    ' - when there is a "new" UserID/Password, at which point it can be retrieved.
    Private Const UserIdPasswordRequired As Boolean = True


    ' *****************************************************************************************************
    ' Public Variables
    ' *****************************************************************************************************

    Public Enum DataPumpMessages
        MSG_NULL = 0
        MSG_UNLOADEXE
        MSG_USERIDPASSWORD
        MSG_REMOVETICKER
        MSG_ADDTICKER
    End Enum

    Public UserID As String    ' Datafeed UserId set in NeuroShell Trader (if UserIdPasswordRequired=True)
    Public Password As String  ' Datafeed Password set in NeuroShell Trader (if UserIdPasswordRequired=True)

    Public AdjustmentToUSEasternTimeZone As Double ' All timestamps fed to the data pump must be set to U.S. Eastern Time Zone.  Use this variable to adjust from the current time zone (if that is how your data arrives)

    ' Arrays to keep track of which ticker symbol/barinterval combinations are already being tracked by the data pump/data feed app
    Public NumberTickers As Integer = 0
    Public TickerSymbol(500) As String
    Public TickerBarInterval(500) As Single
    Public TickerDescription(500) As String
    Public TickerCategory(500) As String
    Public TickerStartDate(500) As String
    Public TickerRealTimeUpdatesHold(500) As Boolean

    Public TickerFxcmInterval(500) As String
    Public TickerLastDateTime(500) As Date
    Public TickerLastDateTimeCount(500) As Integer
    Public TickerLastVolume(500) As Double
    Public TickerUpdate(500) As Boolean

    ' Variables to store list of ticks that occured during the historical data collection (so ticks aren't missed)
    Private InProgressHistoricalTicker As String = ""
    Private InProgressHistoricalBarInterval As Single
    Private InProgressHistoricalTickCount As Integer = 0
    Private InProgressHistoricalTickAllocated As Integer = 0
    Private InProgressHistoricalTickDateTime(0) As Double
    Private InProgressHistoricalTickPrice(0) As Single
    Private InProgressHistoricalTickVolume(0) As Single

    ' *****************************************************************************************************
    ' Data Pump DLL Declares
    ' *****************************************************************************************************

    Private Declare Sub SetDataFeedhwnd Lib "NSTFEED.DLL" (ByVal hWnd As Integer)
    Private Declare Sub RequireUserIDPassword Lib "NSTFEED.DLL" ()
    Private Declare Sub GetUserIDPassword Lib "NSTFEED.DLL" (ByRef UserID As String, ByRef Password As String)
    Private Declare Function NSTSETUPDATA Lib "NSTFEED.DLL" (ByVal Ticker As String, ByVal Description As String, ByVal BarInt As Single, ByVal NumColumns As Integer, ByVal Category As String, ByVal FirstDateAvailable As Double) As Integer
    Private Declare Function NSTSETLABEL Lib "NSTFEED.DLL" (ByVal Ticker As String, ByVal BarInt As Single, ByVal Column As Integer, ByVal LabelName As String) As Integer
    Private Declare Function NSTSETVALUE Lib "NSTFEED.DLL" (ByVal Ticker As String, ByVal BarInt As Single, ByVal Column As Integer, ByVal Value As Double) As Integer
    Private Declare Function NSTCOMMITROW Lib "NSTFEED.DLL" (ByVal Ticker As String, ByVal BarInt As Single) As Integer
    Public Declare Sub UpdatePercent2 Lib "NSTFEED.DLL" (ByVal percent As Integer)

    Public Sub DataPump_Connect(ByVal TextHwnd As Integer)
        ' ********************************************
        ' MUST BE CALLED WHEN APPLICATION FIRST STARTS
        ' ********************************************

        ' Send NST Data Pump the handle to the text box that data pump will use to send events - 
        SetDataFeedhwnd(TextHwnd)

        ' Let NST Data Pump know if a UserID/Password is required to be entered by user in NST
        If UserIdPasswordRequired Then RequireUserIDPassword()
    End Sub

    Public Sub DataPump_Disconnect()
        ' ********************************************
        ' MUST BE CALLED BEFORE APPLICATION CLOSES
        ' ********************************************

        ' Clear the handle to the text box that data pump will use to send events, so the NST Data Pump doesn't keep trying to send events
        SetDataFeedhwnd(0)
    End Sub

#If FULLBIDASKDATA > 0 Then
    Public Sub DataPump_AddHistoricalBar(ByRef Ticker As String, ByRef BarInterval As Single, ByRef DateTimeVal As Double, ByRef OpenVal As Double, ByRef HighVal As Double, ByRef LowVal As Double, ByRef CloseVal As Double, ByRef VolumeVal As Double, ByRef BidOpenVal As Double, ByRef BidHighVal As Double, ByRef BidLowVal As Double, ByRef BidCloseVal As Double, ByRef AskOpenVal As Double, ByRef AskHighVal As Double, ByRef AskLowVal As Double, ByRef AskCloseVal As Double)
#Else
    Public Sub DataPump_AddHistoricalBar(ByRef Ticker As String, ByRef BarInterval As Single, ByRef DateTimeVal As Double, ByRef OpenVal As Double, ByRef HighVal As Double, ByRef LowVal As Double, ByRef CloseVal As Double, ByRef VolumeVal As Double)
#End If
        ' Adds a historical OHLCV bar 
        ' NOTE: must be called from least recent date up to most recent date

        Try
            If BarInterval <= 0 Then
                ' send OHLC data to data pump (but send each piece one by one as individual "ticks" since barinterval<=0 and tick data expected)
                NSTSETVALUE(Ticker, BarInterval, 1, DateTimeVal) 'Date
                NSTSETVALUE(Ticker, BarInterval, 2, OpenVal) 'Open value
                NSTSETVALUE(Ticker, BarInterval, 3, 0) '0 Volume value
                NSTCOMMITROW(Ticker, BarInterval)

                NSTSETVALUE(Ticker, BarInterval, 1, DateTimeVal) 'Date
                NSTSETVALUE(Ticker, BarInterval, 2, HighVal) 'High value
                NSTSETVALUE(Ticker, BarInterval, 3, 0) '0 Volume value
                NSTCOMMITROW(Ticker, BarInterval)

                NSTSETVALUE(Ticker, BarInterval, 1, DateTimeVal) 'Date
                NSTSETVALUE(Ticker, BarInterval, 2, LowVal) 'Low value
                NSTSETVALUE(Ticker, BarInterval, 3, 0) '0 Volume value
                NSTCOMMITROW(Ticker, BarInterval)

                NSTSETVALUE(Ticker, BarInterval, 1, DateTimeVal) 'Date
                NSTSETVALUE(Ticker, BarInterval, 2, CloseVal) 'Close value
                NSTSETVALUE(Ticker, BarInterval, 3, VolumeVal) 'Volume value
                NSTCOMMITROW(Ticker, BarInterval)
            Else
                'MsgBox("Adding data point " + Ticker + ", " + CStr(BarInterval) + ", " + CStr(CDate(Date.FromOADate(DateTimeVal))) + ", " + CStr(OpenVal) + ", " + CStr(HighVal) + ", " + CStr(LowVal) + ", " + CStr(CloseVal) + ", " + CStr(VolumeVal))
                ' Send data as bar (i.e. DateTime / Open / High / Low/ Close / Volume )
                NSTSETVALUE(Ticker, BarInterval, 1, DateTimeVal) 'Date
                NSTSETVALUE(Ticker, BarInterval, 2, OpenVal) 'Open
                NSTSETVALUE(Ticker, BarInterval, 3, HighVal) 'High
                NSTSETVALUE(Ticker, BarInterval, 4, LowVal) 'Low
                NSTSETVALUE(Ticker, BarInterval, 5, CloseVal) 'Close
                NSTSETVALUE(Ticker, BarInterval, 6, VolumeVal) 'Volume
#If FULLBIDASKDATA = 1 Then
                NSTSETVALUE(Ticker, BarInterval, 7, BidCloseVal) 'Close
                NSTSETVALUE(Ticker, BarInterval, 8, AskCloseVal) 'Close
#ElseIf FULLBIDASKDATA = 2 Then
                NSTSETVALUE(Ticker, BarInterval, 7, BidOpenVal) 'Open
                NSTSETVALUE(Ticker, BarInterval, 8, BidHighVal) 'High
                NSTSETVALUE(Ticker, BarInterval, 9, BidLowVal) 'Low
                NSTSETVALUE(Ticker, BarInterval, 10, BidCloseVal) 'Close
                NSTSETVALUE(Ticker, BarInterval, 11, AskOpenVal) 'Open
                NSTSETVALUE(Ticker, BarInterval, 12, AskHighVal) 'High
                NSTSETVALUE(Ticker, BarInterval, 13, AskLowVal) 'Low
                NSTSETVALUE(Ticker, BarInterval, 14, AskCloseVal) 'Close
#End If
                NSTCOMMITROW(Ticker, BarInterval)
            End If
        Catch e As Exception
            MsgBox("DataPump_AddHistoricalBar Exception: " + e.ToString())
            MsgBox("DataPump_AddHistoricalBar Exception: " + Ticker + " BarInterval=" + CStr(BarInterval) + " DateTime=" + CStr(DateTimeVal) + " Open=" + CStr(OpenVal) + " High=" + CStr(HighVal) + " Low=" + CStr(LowVal) + " Close=" + CStr(CloseVal) + " Volume=" + CStr(VolumeVal))
        End Try

    End Sub

#If FULLBIDASKDATA > 0 Then
    Public Sub DataPump_UpdateRealTimeBar(ByRef Ticker As String, ByRef BarInterval As Single, ByRef BarDateTimeStamp As Double, ByRef OpenVal As Double, ByRef HighVal As Double, ByRef LowVal As Double, ByRef CloseVal As Double, ByRef VolVal As Double, ByRef BidOpenVal As Double, ByRef BidHighVal As Double, ByRef BidLowVal As Double, ByRef BidCloseVal As Double, ByRef AskOpenVal As Double, ByRef AskHighVal As Double, ByRef AskLowVal As Double, ByRef AskCloseVal As Double)
#Else
    Public Sub DataPump_UpdateRealTimeBar(ByRef Ticker As String, ByRef BarInterval As Single, ByRef BarDateTimeStamp As Double, ByRef OpenVal As Double, ByRef HighVal As Double, ByRef LowVal As Double, ByRef CloseVal As Double, ByRef VolVal As Double)
#End If
        ' Sends the current OHLC for the current bar to data pump'
        ' NOTE: Used only when actual real time price ticks are unavailable and instead polling for most recent OHLC bar values are used

        Try
            If BarInterval > 0 Then
                NSTSETVALUE(Ticker, BarInterval, 1, BarDateTimeStamp)
                NSTSETVALUE(Ticker, BarInterval, 2, OpenVal)
                NSTSETVALUE(Ticker, BarInterval, 3, HighVal)
                NSTSETVALUE(Ticker, BarInterval, 4, LowVal)
                NSTSETVALUE(Ticker, BarInterval, 5, CloseVal)
                NSTSETVALUE(Ticker, BarInterval, 6, VolVal)
#If FULLBIDASKDATA = 1 Then
                NSTSETVALUE(Ticker, BarInterval, 7, BidCloseVal) 'Close
                NSTSETVALUE(Ticker, BarInterval, 8, AskCloseVal) 'Close
#ElseIf FULLBIDASKDATA = 2 Then
                NSTSETVALUE(Ticker, BarInterval, 7, BidOpenVal) 'Open
                NSTSETVALUE(Ticker, BarInterval, 8, BidHighVal) 'High
                NSTSETVALUE(Ticker, BarInterval, 9, BidLowVal) 'Low
                NSTSETVALUE(Ticker, BarInterval, 10, BidCloseVal) 'Close
                NSTSETVALUE(Ticker, BarInterval, 11, AskOpenVal) 'Open
                NSTSETVALUE(Ticker, BarInterval, 12, AskHighVal) 'High
                NSTSETVALUE(Ticker, BarInterval, 13, AskLowVal) 'Low
                NSTSETVALUE(Ticker, BarInterval, 14, AskCloseVal) 'Close
#End If
                NSTCOMMITROW(Ticker, BarInterval)
            Else
                MsgBox("Programming Error - Bar Interval (" + CStr(BarInterval) + ") is less than or equal to zero, so either change the DataServerType constant to 1 (or 0) or use DataPump_AddPriceTick instead of DataPump_UpdateRealTimeBar")
            End If
        Catch e As Exception
            MsgBox("DataPump_AddHistoricalBar Exception: " + Ticker + " BarInterval=" + CStr(BarInterval) + " DateTime=" + CStr(Convert.ToDateTime(BarDateTimeStamp)) + " Open=" + CStr(OpenVal) + " High=" + CStr(HighVal) + " Low=" + CStr(LowVal) + " Close=" + CStr(CloseVal) + " Volume=" + CStr(VolVal))
        End Try
    End Sub

    Public Sub DataPump_NewPriceTick(ByRef Ticker As String, ByRef TickDateTimeStamp As Double, ByRef PriceVal As Double, ByRef VolVal As Double)
        ' Adds new tick to all appropriate ticker symbol/barinterval combinations that are being tracked by the data pump/data feed app 
        ' If a historical history load is in progress, keeps a running list of ticks for that particular ticker/barinterval so no ticks are missed and they can be added in later after historical load is done

        Dim i As Integer

        If Ticker = InProgressHistoricalTicker Then
            ' update inprogress historical tick buffer
            If InProgressHistoricalTickCount + 1 >= InProgressHistoricalTickAllocated Then
                InProgressHistoricalTickAllocated = InProgressHistoricalTickAllocated + 100
                ReDim Preserve InProgressHistoricalTickDateTime(InProgressHistoricalTickAllocated)
                ReDim Preserve InProgressHistoricalTickPrice(InProgressHistoricalTickAllocated)
                ReDim Preserve InProgressHistoricalTickVolume(InProgressHistoricalTickAllocated)
            End If
            InProgressHistoricalTickDateTime(InProgressHistoricalTickCount) = TickDateTimeStamp
            InProgressHistoricalTickPrice(InProgressHistoricalTickCount) = PriceVal
            InProgressHistoricalTickVolume(InProgressHistoricalTickCount) = VolVal
            InProgressHistoricalTickCount = InProgressHistoricalTickCount + 1

            ' look for symbol and bar interval matches (BUT exclude currently in progress historical)
            For i = 1 To NumberTickers
                ' check for symbol and bar interval match
                If TickerSymbol(i) = Ticker And TickerBarInterval(i) <> InProgressHistoricalBarInterval Then
                    DataPump_AddPriceTick(Ticker, TickerBarInterval(i), TickDateTimeStamp, PriceVal, VolVal)
                End If
            Next i
        Else
            ' look for symbol and bar interval matches
            For i = 1 To NumberTickers
                ' check for symbol and bar interval match
                If TickerSymbol(i) = Ticker Then
                    DataPump_AddPriceTick(Ticker, TickerBarInterval(i), TickDateTimeStamp, PriceVal, VolVal)
                End If
            Next i
        End If
    End Sub

    Public Sub DataPump_AddPriceTick(ByRef Ticker As String, ByRef BarInterval As Single, ByRef TickDateTimeStamp As Double, ByRef PriceVal As Double, ByRef VolVal As Double)
        ' Sends price ticks for historical bars (BarInterval=0) and realtime ticks (BarInterval=0 and BarInterval<0)

        Try
            If BarInterval <= 0 Then
                'MsgBox("DataPump_AddPriceTick " + Ticker + " " + CStr(BarInterval) + " " + CStr(TickDateTimeStamp) + " " + CStr(PriceVal) + " " + CStr(VolVal))
                If PriceVal <= 0 Then
                    'MsgBox("DataPump_AddPriceTick price is zero (or negative)")
                Else
                    NSTSETVALUE(Ticker, BarInterval, 1, TickDateTimeStamp)
                    NSTSETVALUE(Ticker, BarInterval, 2, PriceVal)
                    NSTSETVALUE(Ticker, BarInterval, 3, VolVal)
                    NSTCOMMITROW(Ticker, BarInterval)
                End If
            Else
                MsgBox("Programming Error - Bar Interval (" + CStr(BarInterval) + ") is greater than zero, so either change the DataServerType constant to 2 or use DataPump_UpdateRealTimeBar instead of DataPump_AddPriceTick")
            End If
        Catch e As Exception
            MsgBox("DataPump_AddHistoricalBar Exception: " + Ticker + " BarInterval=" + CStr(BarInterval) + " DateTime=" + CStr(Convert.ToDateTime(TickDateTimeStamp)) + " Price=" + CStr(PriceVal) + " Volume=" + CStr(VolVal))
        End Try
    End Sub

    Public Function DataPump_ParseString(ByVal Txt As String, ByRef Ticker As String, ByRef BarInterval As Single, ByRef Description As String, ByRef Category As String, ByRef StartDate As String, ByRef Columns As Collection) As Integer
        ' Parses strings sent to the text box by the NST data pump

        Dim ColumnNames As String
        Dim BarSize As Integer
        Dim TempBarSize As String
        Dim TempUserID As String = Space$(50)
        Dim TempPassword As String = Space$(50)

        DataPump_ParseString = DataPumpMessages.MSG_NULL

        If Txt <> "" Then
            If Txt = "*****UnloadEXE*****" Then
                DataPump_ParseString = DataPumpMessages.MSG_UNLOADEXE
            ElseIf Txt = "*****UserIDPassword*****" Then
                'The UserIDPassword has changed use 'GetUserIDPassword' to get the changed UserID/Password
                If UserIdPasswordRequired Then
                    TempUserID = "                               "
                    TempPassword = "                               "

                    GetUserIDPassword(TempUserID, TempPassword)

                    UserID = TempUserID
                    While Asc(Right(UserID, 1)) = 0
                        UserID = Left(UserID, Len(UserID) - 1)
                    End While
                    Password = TempPassword
                    While Asc(Right(Password, 1)) = 0
                        Password = Left(Password, Len(Password) - 1)
                    End While

                    DataPump_ParseString = DataPumpMessages.MSG_USERIDPASSWORD
                End If
            ElseIf InStr(Txt, "*****RemoveTicker*****") Then
                Ticker = Right(Txt, Len(Txt) - InStr(Txt, "!@#") - 2)
                TempBarSize = Right(Ticker, Len(Ticker) - InStr(Ticker, "!@#") - 2)

                Ticker = Trim(Left(Ticker, InStr(Ticker, "!@#") - 1))
                BarInterval = Val(TempBarSize)

                DataPump_ParseString = DataPumpMessages.MSG_REMOVETICKER
            ElseIf InStr(Txt, "!@#") Then
                ' Parse out ticker, description, category, barinterval, startdate and column names
                Ticker = Trim(Left(Txt, InStr(Txt, "!@#") - 1))
                Description = Right(Txt, Len(Txt) - InStr(Txt, "!@#") - 2)
                Category = Right(Description, Len(Description) - InStr(Description, "!@#") - 2)
                TempBarSize = Right(Category, Len(Category) - InStr(Category, "!@#") - 2)
                StartDate = Right(TempBarSize, Len(TempBarSize) - InStr(TempBarSize, "!@#") - 2)
                ColumnNames = Right(StartDate, Len(StartDate) - InStr(StartDate, "!@#") - 2)

                Description = Trim(Left(Description, InStr(Description, "!@#") - 1))
                Category = Trim(Left(Category, InStr(Category, "!@#") - 1))
                TempBarSize = Trim(Left(TempBarSize, InStr(TempBarSize, "!@#") - 1))
                BarSize = Val(TempBarSize)
                StartDate = Trim(Left(StartDate, InStr(StartDate, "!@#") - 1))

                ' Set BarInterval based upon BarSize requested and the way your data server provides minute/hour/daily data
                If BarSize = 0 Then
                    ' Tick by tick data being requested
                    BarInterval = 0
                ElseIf DataServerType = 0 Then
                    BarInterval = 0 ' tick by tick historical and realtime updates for minute/hour bars
                ElseIf DataServerType = 1 Then
                    BarInterval = -BarSize ' OHLC historical and tick by tick realtime updates for minute/hour bars
                ElseIf DataServerType = 2 Then
                    BarInterval = BarSize ' OHLC historical and OHLC realtime polling/updates for minute/hour bars
                End If

                While InStr(ColumnNames, ",")
                    Columns.Add(Left(ColumnNames, InStr(ColumnNames, ",") - 1))
                    ColumnNames = Right(ColumnNames, Len(ColumnNames) - InStr(ColumnNames, ","))
                End While
                Columns.Add(ColumnNames)

                DataPump_ParseString = DataPumpMessages.MSG_ADDTICKER
            End If
        End If
    End Function

    Public Sub DataPump_AddTickerSymbol(ByRef FirstBarAvailableDateTime As Double, ByRef Ticker As String, ByRef BarInterval As Single, ByRef Description As String, ByRef Category As String) ', ByRef Columns As Collection)
        ' Adds a new ticker symbol/barinterval combination to the list of ticker symbols being tracked by the data pump/data feed app

        Dim temp As Integer
        Dim ticknumber As Integer

        ticknumber = DataPump_TickerNumberLookup(Ticker, BarInterval)
        If ticknumber = 0 Then
            ' increase number of data streams
            NumberTickers = NumberTickers + 1
        End If

        ' set values
        TickerSymbol(NumberTickers) = Ticker
        TickerBarInterval(NumberTickers) = BarInterval
        TickerDescription(NumberTickers) = Description
        TickerCategory(NumberTickers) = Category
        TickerStartDate(NumberTickers) = CStr(FirstBarAvailableDateTime)

        ' Setup data pump with new ticker, bar interval, data labels, etc
        If BarInterval = 0 Then
            temp = NSTSETUPDATA(Ticker, Description, BarInterval, 3, Category, CDbl(Fix(FirstBarAvailableDateTime)))
            temp = NSTSETLABEL(Ticker, BarInterval, 1, "Date/Time")
            temp = NSTSETLABEL(Ticker, BarInterval, 2, "Price")
            temp = NSTSETLABEL(Ticker, BarInterval, 3, "Volume")
            ' Can put in trade type for bid/ask stuff as well
        ElseIf BarInterval < 0 Then
            temp = NSTSETUPDATA(Ticker, Description, BarInterval, 3, Category, CDbl(Fix(FirstBarAvailableDateTime)))
            temp = NSTSETLABEL(Ticker, BarInterval, 1, "Date/Time")
            temp = NSTSETLABEL(Ticker, BarInterval, 2, "Price")
            temp = NSTSETLABEL(Ticker, BarInterval, 3, "Volume")
        ElseIf BarInterval > 0 Then
#If FULLBIDASKDATA = 1 Then
            temp = NSTSETUPDATA(Ticker, Description, BarInterval, 8, Category, CDbl(Fix(FirstBarAvailableDateTime)))
            temp = NSTSETLABEL(Ticker, BarInterval, 1, "Date/Time")
            temp = NSTSETLABEL(Ticker, BarInterval, 2, "Open")
            temp = NSTSETLABEL(Ticker, BarInterval, 3, "High")
            temp = NSTSETLABEL(Ticker, BarInterval, 4, "Low")
            temp = NSTSETLABEL(Ticker, BarInterval, 5, "Close")
            temp = NSTSETLABEL(Ticker, BarInterval, 6, "Volume")
            temp = NSTSETLABEL(Ticker, BarInterval, 7, "Bid")
            temp = NSTSETLABEL(Ticker, BarInterval, 8, "Ask")
#ElseIf FULLBIDASKDATA = 2 Then
            temp = NSTSETUPDATA(Ticker, Description, BarInterval, 14, Category, CDbl(Fix(FirstBarAvailableDateTime)))
            temp = NSTSETLABEL(Ticker, BarInterval, 1, "Date/Time")
            temp = NSTSETLABEL(Ticker, BarInterval, 2, "Open")
            temp = NSTSETLABEL(Ticker, BarInterval, 3, "High")
            temp = NSTSETLABEL(Ticker, BarInterval, 4, "Low")
            temp = NSTSETLABEL(Ticker, BarInterval, 5, "Close")
            temp = NSTSETLABEL(Ticker, BarInterval, 6, "Volume")
            temp = NSTSETLABEL(Ticker, BarInterval, 7, "Bid Open")
            temp = NSTSETLABEL(Ticker, BarInterval, 8, "Bid High")
            temp = NSTSETLABEL(Ticker, BarInterval, 9, "Bid Low")
            temp = NSTSETLABEL(Ticker, BarInterval, 10, "Bid Close")
            temp = NSTSETLABEL(Ticker, BarInterval, 11, "Ask Open")
            temp = NSTSETLABEL(Ticker, BarInterval, 12, "Ask High")
            temp = NSTSETLABEL(Ticker, BarInterval, 13, "Ask Low")
            temp = NSTSETLABEL(Ticker, BarInterval, 14, "Ask Close")
#Else
            temp = NSTSETUPDATA(Ticker, Description, BarInterval, 6, Category, CDbl(Fix(FirstBarAvailableDateTime)))
            temp = NSTSETLABEL(Ticker, BarInterval, 1, "Date/Time")
            temp = NSTSETLABEL(Ticker, BarInterval, 2, "Open")
            temp = NSTSETLABEL(Ticker, BarInterval, 3, "High")
            temp = NSTSETLABEL(Ticker, BarInterval, 4, "Low")
            temp = NSTSETLABEL(Ticker, BarInterval, 5, "Close")
            temp = NSTSETLABEL(Ticker, BarInterval, 6, "Volume")
#End If

        End If

        ' Can add additional columns beyond just OHLCV if needed
        'For i = 1 To Columns.Count
        '    temp = NSTSETLABEL(Ticker, BarInterval, i, Columns(i))
        'Next i
    End Sub

    Public Function DataPump_RemoveTicker(ByRef Ticker As String, ByRef BarInterval As Single) As Integer
        ' Removes a new ticker symbol/barinterval combination from the list of ticker symbols being tracked by the data pump/data feed app

        Dim j As Integer
        Dim TickerNumber As Integer

        TickerNumber = DataPump_TickerNumberLookup(Ticker, BarInterval)
        If TickerNumber > 0 Then
            ' reduce number of tickers
            NumberTickers = NumberTickers - 1

            ' shift all higher tickers down by one
            For j = TickerNumber To NumberTickers
                TickerSymbol(j) = TickerSymbol(j + 1)
                TickerBarInterval(j) = TickerBarInterval(j + 1)
                TickerDescription(j) = TickerDescription(j + 1)
                TickerCategory(j) = TickerCategory(j + 1)
                TickerStartDate(j) = TickerStartDate(j + 1)
                TickerRealTimeUpdatesHold(j) = TickerRealTimeUpdatesHold(j + 1)
                TickerFxcmInterval(j) = TickerFxcmInterval(j + 1)
                TickerLastDateTime(j) = TickerLastDateTime(j + 1)
                TickerLastDateTimeCount(j) = TickerLastDateTimeCount(j + 1)
                TickerLastVolume(j) = TickerLastVolume(j + 1)
                TickerUpdate(j) = TickerUpdate(j + 1)
            Next j
        End If

        DataPump_RemoveTicker = TickerNumber
    End Function

    Public Function DataPump_TickerNumberLookup(ByRef Ticker As String, ByRef BarInterval As Single) As Integer
        ' Checks to see if a ticker symbol/barinterval combination is already being tracked by the data pump/data feed app

        Dim i As Integer

        ' look for symbol and bar interval match
        DataPump_TickerNumberLookup = 0 ' default to zero for not found
        For i = 1 To NumberTickers
            ' check for symbol and bar interval match
            If TickerSymbol(i) = Ticker And TickerBarInterval(i) = BarInterval Then
                DataPump_TickerNumberLookup = i
                Exit For
            End If
        Next i
    End Function

    Public Sub DataPump_HistoricalTickBufferStart(ByRef Ticker As String, ByRef BarInterval As Single)
        ' buffers incoming realtime ticks while a historical data request is being done in hopes of minimizing the number of ticks that are lost or overlapped between the historical bars and incoming realtime ticks.

        ' For example, if historical data is requested from the server from most recent to least recent date and the request takes 30 seconds, there are potentialy 30 seconds of ticks that occured since the most recent
        ' historical bar's data was sent from the server.  However, those ticks can't be added to the data pump until after the historical data is put into the data pump (which can only be done after the least recent bar is collected
        ' since the data pump requires data bars to be filled from least recent to most recent).  So those ticks need to be buffered and then added after the historical bars have been added.

        InProgressHistoricalTicker = Ticker
        InProgressHistoricalBarInterval = BarInterval
        InProgressHistoricalTickCount = 0
    End Sub

    Public Sub DataPump_HistoricalTickBufferEnd(ByRef LastBarDateTimeVal As Double, ByRef LastBarOpenVal As Double, ByRef LastBarHighVal As Double, ByRef LastBarLowVal As Double, ByRef LastBarCloseVal As Double, ByRef LastBarVolumeVal As Double, LastHistoricalTickDateTime As Double, LastHistoricalTickVolume As Double)
        ' Adds ticks buffered during historical data request
        ' and stops the future buffering of incoming the realtime ticks (ticks will just be added as they arrive after code below executes)

        Dim flag As Boolean = False
        Dim FirstFlag As Boolean = True

        ' If LastHistoricalTickDateTime = 0, then all buffered ticks are added to data stream (could result in overlap of ticks included in last collected historical bars and buffered ticks depending upon timing)
        ' If LastHistoricalTickDateTime set, but LastHistoricalTickVolume = 0, then all buffered ticks after LastHistoricalTickDateTime are added to data stream (could result in mising or overlap of ticks in the last second of the collected historical bars and buffered ticks depending upon timing)
        ' If LastHistoricalTickDateTime and LastHistoricalTickVolume set, but then all buffered ticks after the tick with a value the same as LastHistoricalTickDateTime and LastHistoricalTickVolume are added to data stream (could result in missing or overlap of ticks if multiple ticks in same second with same volume or if the exact tick was not buffered for some reason)

        ' Add last bar's open, high, low and last close value as ticks to be combined with ticks collected while historical was being requested (Note: LastBarDateTimeVal should be actual time of when last bar was collected and not when the last bar will end in the future for best combining with ticks)
        'MsgBox("HistoricalTickBufferEnd adding last bar (LastBarVolumeVal=" + CStr(LastBarVolumeVal) + "  LastHistoricalTickVolume=" + CStr(LastHistoricalTickVolume) + ") and evaluating " + CStr(InProgressHistoricalTickCount) + " ticks")
        DataPump_AddPriceTick(InProgressHistoricalTicker, InProgressHistoricalBarInterval, LastBarDateTimeVal, LastBarOpenVal, 0)
        DataPump_AddPriceTick(InProgressHistoricalTicker, InProgressHistoricalBarInterval, LastBarDateTimeVal, LastBarHighVal, 0)
        DataPump_AddPriceTick(InProgressHistoricalTicker, InProgressHistoricalBarInterval, LastBarDateTimeVal, LastBarLowVal, 0)
        DataPump_AddPriceTick(InProgressHistoricalTicker, InProgressHistoricalBarInterval, LastBarDateTimeVal, LastBarCloseVal, LastBarVolumeVal)

        ' For fast processing, split out tick matching based upon the amount of information available on the last tick included in the collected historical bars
        If LastHistoricalTickDateTime = 0 Then
            ' add all buffered ticks to data pump 
            For i As Integer = 0 To InProgressHistoricalTickCount - 1
                DataPump_AddPriceTick(InProgressHistoricalTicker, InProgressHistoricalBarInterval, InProgressHistoricalTickDateTime(i), InProgressHistoricalTickPrice(i), InProgressHistoricalTickVolume(i))
                If FirstFlag Then
                    'MsgBox("HistoricalTickBufferEnd   Add All Ticks ->  BarStopTime=0  /  TickTime=" + CStr(InProgressHistoricalTickDateTime(i)))
                    FirstFlag = False
                End If
            Next i
        ElseIf LastHistoricalTickVolume = 0 Then
            ' add all buffered ticks after the LastHistoricalTickDateTime (i.e. the date/time of the last tick that was included in the historical bar history request)
            For i As Integer = 0 To InProgressHistoricalTickCount - 1
                If InProgressHistoricalTickDateTime(i) > LastHistoricalTickDateTime Then
                    DataPump_AddPriceTick(InProgressHistoricalTicker, InProgressHistoricalBarInterval, InProgressHistoricalTickDateTime(i), InProgressHistoricalTickPrice(i), InProgressHistoricalTickVolume(i))
                    If FirstFlag Then
                        'MsgBox("HistoricalTickBufferEnd   TimeMatch ->  BarStopTime=" + CStr(LastHistoricalTickDateTime) + "  /  TickTime=" + CStr(InProgressHistoricalTickDateTime(i)) + "   LastHistoricalTickVolume=" + Format$(LastHistoricalTickVolume))
                        FirstFlag = False
                    End If
                End If
            Next i
        Else
            ' add all buffered ticks after the tick with the a date time of LastHistoricalTickDateTime and a Volume of LastHistoricalTickVolume
            For i As Integer = 0 To InProgressHistoricalTickCount - 1
                ' check if last tick included in historical bars has been passed based upon date/time (but not caught by volume for some reason ... missing ticks etc.)
                If Not flag Then
                    flag = InProgressHistoricalTickDateTime(i) > LastHistoricalTickDateTime
                    If FirstFlag Then
                        'MsgBox("HistoricalTickBufferEnd   TimeAndVolumeMatch, but Volume Not found ->  BarStopTime=" + CStr(LastHistoricalTickDateTime) + "  /  TickTime=" + CStr(InProgressHistoricalTickDateTime(i)) + "   LastHistoricalTickVolume=" + Format$(LastHistoricalTickVolume) + "  CurrentTickVolume=" + Format$(InProgressHistoricalTickVolume(i)))
                        FirstFlag = False
                    End If
                End If

                ' start adding buffered ticks once the last tick included in historical bars is passed
                If flag Then
                    DataPump_AddPriceTick(InProgressHistoricalTicker, InProgressHistoricalBarInterval, InProgressHistoricalTickDateTime(i), InProgressHistoricalTickPrice(i), InProgressHistoricalTickVolume(i))
                End If

                ' check if this tick is the last tick included in the historical bars (based upon time stamp and volume value)
                If Not flag Then
                    flag = (InProgressHistoricalTickDateTime(i) = LastHistoricalTickDateTime And InProgressHistoricalTickVolume(i) = LastHistoricalTickVolume)
                    If FirstFlag Then
                        'MsgBox("HistoricalTickBufferEnd   TimeAndVolumeMatch ->  BarStopTime=" + CStr(LastHistoricalTickDateTime) + "  /  OverlappingTickTime=" + CStr(InProgressHistoricalTickDateTime(i)) + " OverlappingTickVol=" + CStr(LastHistoricalTickVolume) + "   LastHistoricalTickVolume=" + Format$(LastHistoricalTickVolume) + "  CurrentTickVolume=" + Format$(InProgressHistoricalTickVolume(i)))
                        FirstFlag = False
                    End If
                End If
            Next i

        End If

        ' clear inprogresshistorical variables (so new ticks for this ticker/barinterval will start being added as they come in instead of being buffered)
        InProgressHistoricalTicker = ""
        InProgressHistoricalBarInterval = 0

        ' deallocate array memory
        InProgressHistoricalTickCount = 0
        InProgressHistoricalTickAllocated = 0
        ReDim InProgressHistoricalTickDateTime(0)
        ReDim InProgressHistoricalTickPrice(0)
        ReDim InProgressHistoricalTickVolume(0)
    End Sub
End Module